Package: ebal
Type: Package
Title: Entropy Reweighting to Create Balanced Samples
Version: 0.2.1
Date: 2026-04-28
Authors@R: person("Jens", "Hainmueller", email = "jhain@stanford.edu",
                  role = c("aut", "cre"))
Description: Implements entropy balancing, a data preprocessing procedure
    described in Hainmueller (2012, <doi:10.1093/pan/mpr025>) that allows
    users to reweight a dataset such that the covariate distributions in
    the reweighted data satisfy a set of user-specified moment conditions.
    Useful for creating balanced samples in observational studies with a
    binary treatment where the control group is reweighted to match the
    covariate moments of the treatment group, and for reweighting a
    survey sample to known characteristics from a target population.
License: GPL (>= 2)
Imports: graphics, methods, stats
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
URL: https://web.stanford.edu/~jhain/, https://github.com/j-hai/ebal
BugReports: https://github.com/j-hai/ebal/issues
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-04-28 19:04:18 UTC; jhainmueller
Author: Jens Hainmueller [aut, cre]
Maintainer: Jens Hainmueller <jhain@stanford.edu>
Repository: CRAN
Date/Publication: 2026-04-29 07:30:08 UTC
Built: R 4.5.2; ; 2026-05-01 17:09:07 UTC; unix

importFrom("stats", "model.frame", "model.matrix", "model.response",
           "na.pass", "optimize", "var", "weighted.mean", "weights")
importFrom("graphics", "abline", "axis", "legend", "par", "points")
importFrom("methods", "is")

export(eb,
       ebalance,
       ebalance.trim,
       baltest.collect,
       getsquares,
       line.searcher,
       matrixmaker)

# S3 method registrations for the print/summary/plot/weights methods on
# ebalance and ebalance.trim objects. We intentionally do NOT make
# ebalance() itself an S3 generic: the formula vs matrix dispatch is
# handled in-function by inherits(Treatment, "formula"). This avoids a
# CRAN R CMD check NOTE that would otherwise flag the long-standing
# top-level function ebalance.trim() as an "apparent method" of a
# (would-be) generic ebalance().
S3method(print, ebalance)
S3method(print, ebalance.trim)
S3method(print, summary.ebalance)
S3method(print, summary.ebalance.trim)
S3method(summary, ebalance)
S3method(summary, ebalance.trim)
S3method(plot, ebalance)
S3method(plot, ebalance.trim)
S3method(weights, ebalance)
S3method(weights, ebalance.trim)

.onAttach <- function(libname, pkgname) {
  packageStartupMessage("##\n## ebal Package: Implements Entropy Balancing.\n")
  packageStartupMessage("## See https://web.stanford.edu/~jhain/ for additional information.\n")
}

eb <-
function(tr.total=tr.total,
               co.x=co.x,
               coefs=coefs,
               base.weight=base.weight,
               max.iterations=max.iterations,
               constraint.tolerance=constraint.tolerance,
               print.level=print.level
               ) {

converged <- FALSE
for(iter in 1:max.iterations) {
 weights.temp <-  c(.safe_exp(co.x %*% coefs))
 weights.ebal <- weights.temp *  base.weight
 co.x.agg   <- c(weights.ebal %*% co.x)
 gradient   <- co.x.agg - tr.total
 if(max(abs(gradient))<constraint.tolerance){
 converged <- TRUE
 break
 }
 if(print.level>=2){ cat("Iteration",iter,"maximum deviation is =",format(max(abs(gradient)),digits=4),"\n") }
 hessian = t(co.x) %*% (weights.ebal * co.x)
 Coefs  <- coefs
 newton <- solve(hessian,gradient)
 coefs  <- coefs - newton
 loss.new <- line.searcher(Base.weight=base.weight,Co.x=co.x,Tr.total=tr.total,coefs=Coefs,Newton=newton,ss=1)
 loss.old <- line.searcher(Base.weight=base.weight,Co.x=co.x,Tr.total=tr.total,coefs=Coefs,Newton=newton,ss=0)
if(print.level>=3){cat("new loss",loss.new,"old loss=",loss.old,"\n")}

 if(loss.old <= loss.new){
 ss.out <- optimize(line.searcher,
                lower=.00001,upper=1,maximum=FALSE,
                Base.weight=base.weight,Co.x=co.x,Tr.total=tr.total,coefs=Coefs,Newton=newton)

  if(print.level>=3){cat("LS Step Length is ",ss.out$minimum,"\n")}
  if(print.level>=3){cat("Loss is",ss.out$objective,"\n")}
  coefs <- Coefs - ss.out$minimum * newton
 }
}
if(print.level>=1 && converged){cat("Converged within tolerance \n")}
return(
       list(
            maxdiff=max(abs(gradient)),
            coefs=coefs,
            Weights.ebal=weights.ebal,
            converged=converged
            )
       )
}

# function to conduct line search for optimal step length
line.searcher <- function(Base.weight,
                          Co.x,
                          Tr.total,
                          coefs,
                          Newton,
                          ss) {
  weights.temp <- c(.safe_exp(Co.x %*% (coefs - (ss * Newton))))
  weights.temp <- weights.temp * Base.weight
  Co.x.agg     <- c(weights.temp %*% Co.x)
  maxdiff      <- max(abs(Co.x.agg - Tr.total))
  maxdiff
}

# function to trim weights
ebalance.trim <-
  function(
           ebalanceobj,
           max.weight=NULL,
           min.weight=0,
           max.trim.iterations = 200,
           max.weight.increment=.92,
           min.weight.increment= 1.08,
           print.level=0
           )
    {
    if( is(ebalanceobj,"ebalance")==FALSE ){
     stop("ebalanceobj must be an ebalance object from a call to ebalance()")
    }
    minimization <- FALSE
    if(is.null(max.weight)){
     max.weight <- max(ebalanceobj$w/mean(ebalanceobj$w))
     minimization <- TRUE
    }
    if(length(max.weight) != 1){
      stop("length(max.weight) != 1")
    }
    if(length(min.weight) != 1){
      stop("length(min.weight) != 1")
    }

  # starting setup for trimming
   w.trimming <- 1
   coefs <- ebalanceobj$coefs

 ### Trimming to user supplied max weight or trim once starting from max weight obtained from distribution
   eb.out <- NULL
   trim.feasible <- FALSE
   for (iter.trim in 1:max.trim.iterations) {
    if (!minimization && print.level > 0) {
      cat("Trim iteration", format(iter.trim, digits = 3), "\n")
    }
    next.eb <- tryCatch(
      eb(tr.total = ebalanceobj$target.margins,
         co.x = ebalanceobj$co.xdata,
         coefs = coefs,
         base.weight = ebalanceobj$w * w.trimming,
         max.iterations = ebalanceobj$max.iterations,
         constraint.tolerance = ebalanceobj$constraint.tolerance,
         print.level = print.level),
      error = function(e) e
    )
    if (inherits(next.eb, "error")) {
      if (is.null(eb.out)) {
        # First iteration failed outright; cannot recover.
        stop("eb() failed during trimming: ", conditionMessage(next.eb))
      }
      if (!minimization) {
        warning("Trimming halted at iteration ", iter.trim,
                " (", conditionMessage(next.eb),
                "). Returning the most recent feasible fit; the requested ",
                "max.weight target may be infeasible for this data.",
                call. = FALSE)
      }
      break
    }
    eb.out <- next.eb
    weights.ratio <- eb.out$Weights.ebal / mean(eb.out$Weights.ebal)
    coefs <- eb.out$coefs
    if (max(weights.ratio) <= max.weight && min(weights.ratio) >= min.weight) {
      if (!minimization && print.level > 0) cat("Converged within tolerance \n")
      trim.feasible <- TRUE
      break
    }
    w.trimming <- w.trimming * ifelse(weights.ratio > max.weight,
                                      w.trimming * ((max.weight * max.weight.increment) / weights.ratio),
                                      1)
    if (min.weight > 0) {
      w.trimming <- w.trimming * ifelse(weights.ratio < min.weight,
                                        w.trimming * ((min.weight * min.weight.increment) / weights.ratio),
                                        1)
    }
   }

 ### automated trimming to minimize max weight
   if (minimization) {
     # Automated minimization mode: the "feasible" semantics are different
     # from the explicit-target case. The algorithm intentionally pushes
     # max.weight downward until it can't go further, so the returned
     # result is the best achievable trimming and is considered feasible
     # by definition.
     trim.feasible <- TRUE
     if (print.level > 0) cat("Automated trimming of max weight ratio \n")
     for (iter.max.weight in 1:max.trim.iterations) {
        max.weight.old     <- max.weight
        weights.ratio.old  <- max(eb.out$Weights.ebal / mean(eb.out$Weights.ebal))
        eb.out.old <- eb.out # store old weights
        if (print.level > 0) {
           cat("Trim iteration", format(iter.max.weight, digits = 3),
               "Max Weight Ratio:", format(weights.ratio.old, digits = 4), "\n")
        }
        suppressWarnings(IsError <- try(
            for(iter.trim in 1:max.trim.iterations) {
              eb.out <- eb(tr.total=ebalanceobj$target.margins,
               co.x=ebalanceobj$co.xdata,
               coefs=coefs,
               base.weight=ebalanceobj$w*w.trimming,
               max.iterations=ebalanceobj$max.iterations,
               constraint.tolerance=ebalanceobj$constraint.tolerance,
               print.level=0
               )
             weights.ratio <- eb.out$Weights.ebal/mean(eb.out$Weights.ebal)
             coefs <- eb.out$coefs
             if (max(weights.ratio) <= max.weight && min(weights.ratio) >= min.weight) break
             w.trimming <- w.trimming * ifelse(weights.ratio > max.weight,
                                               w.trimming * ((max.weight * max.weight.increment) / weights.ratio),
                                               1)
             if (min.weight > 0) {
               w.trimming <- w.trimming * ifelse(weights.ratio < min.weight,
                                                 w.trimming * ((min.weight * min.weight.increment) / weights.ratio),
                                                 1)
             }
           }
    ,silent=TRUE)) # end try call
    
    if( is(IsError,"try-error") ) {
       if(print.level >= 2) { cat("no further decrease in max weight ratio \n") }
     break
     }
    if(weights.ratio.old < max(weights.ratio)) {
     if(print.level >= 2) { cat("no further decrease in max weight ratio \n") }
     break
    }
      # exit loop if alog doesn converge
     max.weight <- max.weight*max.weight.increment   # otherwise lower max weight and try again
    # cat("Moving into next minimisation loop with old max weight ratio:",max(weights.ratio),"and new attempted max weight ratio",max.weight,"\n")

   } # end max weight loop
  # play back the results
     eb.out <- eb.out.old
     if (eb.out$converged && print.level > 0) {
        cat("Converged within tolerance \n")
     }
   } # end automated if

z <- list(
          target.margins       = ebalanceobj$target.margins,
          co.xdata             = ebalanceobj$co.xdata,
          w                    = eb.out$Weights.ebal,
          coefs                = eb.out$coefs,
          maxdiff              = eb.out$maxdiff,
          norm.constant        = ebalanceobj$norm.constant,
          constraint.tolerance = ebalanceobj$constraint.tolerance,
          max.iterations       = ebalanceobj$max.iterations,
          base.weight          = ebalanceobj$base.weight,
          converged            = eb.out$converged,
          trim.feasible        = trim.feasible,
          Treatment            = ebalanceobj$Treatment,
          X                    = ebalanceobj$X
    )

class(z) <- "ebalance.trim"
return(z)
    
   }
    

test_that("weights() returns a length-n vector aligned to Treatment", {
  set.seed(20260427L)
  treatment <- c(rep(0, 50), rep(1, 30))
  X <- rbind(replicate(3, rnorm(50, 0)), replicate(3, rnorm(30, 0.5)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)

  w <- weights(fit)
  expect_length(w, 80L)
  expect_true(all(w[treatment == 1] == 1))
  expect_equal(w[treatment == 0], fit$w, tolerance = 1e-15)
})

test_that("weights() works on ebalance.trim objects", {
  set.seed(20260427L)
  treatment <- c(rep(0, 50), rep(1, 30))
  X <- rbind(replicate(3, rnorm(50, 0)), replicate(3, rnorm(30, 0.5)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)
  trimmed <- ebalance.trim(fit, print.level = 0)

  w <- weights(trimmed)
  expect_length(w, 80L)
  expect_true(all(w[treatment == 1] == 1))
})

test_that("print.ebalance shows expected fields", {
  set.seed(20260427L)
  treatment <- c(rep(0, 50), rep(1, 30))
  X <- rbind(replicate(3, rnorm(50, 0)), replicate(3, rnorm(30, 0.5)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)

  out <- capture.output(print(fit))
  expect_true(any(grepl("Entropy balancing", out)))
  expect_true(any(grepl("Treated:", out)))
  expect_true(any(grepl("Converged:", out)))
})

test_that("summary.ebalance returns balance table with expected columns", {
  set.seed(20260427L)
  treatment <- c(rep(0, 50), rep(1, 30))
  X <- rbind(replicate(3, rnorm(50, 0)), replicate(3, rnorm(30, 0.5)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)

  s <- summary(fit)
  expect_s3_class(s, "summary.ebalance")
  expect_s3_class(s$balance, "data.frame")
  expected_cols <- c("mean.Tr", "mean.Co.pre", "mean.Co.post",
                     "diff.pre", "diff.post",
                     "std.diff.pre", "std.diff.post")
  expect_true(all(expected_cols %in% colnames(s$balance)))
  # post-weighting standardized differences should be near zero
  expect_true(all(abs(s$balance$std.diff.post) < 0.05))
})

test_that("summary.ebalance.trim includes trim.feasible", {
  set.seed(20260427L)
  treatment <- c(rep(0, 50), rep(1, 30))
  X <- rbind(replicate(3, rnorm(50, 0)), replicate(3, rnorm(30, 0.5)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)
  trimmed <- ebalance.trim(fit, print.level = 0)

  s <- summary(trimmed)
  expect_s3_class(s, "summary.ebalance.trim")
  expect_false(is.null(s$call.info$trim.feasible))
})

test_that("plot.ebalance writes a non-empty PDF", {
  skip_on_cran()  # plotting may be flaky on minimal CRAN setups
  set.seed(20260427L)
  treatment <- c(rep(0, 50), rep(1, 30))
  X <- rbind(replicate(3, rnorm(50, 0)), replicate(3, rnorm(30, 0.5)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)

  pdf_path <- tempfile(fileext = ".pdf")
  pdf(pdf_path)
  expect_no_error(plot(fit))
  dev.off()
  expect_true(file.exists(pdf_path))
  expect_gt(file.size(pdf_path), 0)
  unlink(pdf_path)
})

test_that("ebalance.trim auto-minimization succeeds", {
  set.seed(20260427L)
  treatment <- c(rep(0, 100), rep(1, 50))
  X <- rbind(replicate(3, rnorm(100, 0)),
             replicate(3, rnorm(50, 0.6)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)

  trimmed <- ebalance.trim(fit, print.level = 0)
  expect_s3_class(trimmed, "ebalance.trim")
  expect_true(trimmed$trim.feasible)
  expect_true(trimmed$converged)
  # auto mode reduces max weight ratio compared to untrimmed fit
  expect_lt(max(trimmed$w / mean(trimmed$w)),
            max(fit$w / mean(fit$w)) + 1e-6)
})

test_that("ebalance.trim with achievable target meets it and reports feasible", {
  set.seed(20260427L)
  treatment <- c(rep(0, 100), rep(1, 50))
  X <- rbind(replicate(3, rnorm(100, 0)),
             replicate(3, rnorm(50, 0.6)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)
  base.ratio <- max(fit$w / mean(fit$w))
  target <- base.ratio * 1.5

  trimmed <- ebalance.trim(fit, max.weight = target, print.level = 0)
  expect_true(trimmed$trim.feasible)
  expect_lte(max(trimmed$w / mean(trimmed$w)), target)
})

test_that("ebalance.trim with infeasible target falls back gracefully", {
  set.seed(20260427L)
  treatment <- c(rep(0, 100), rep(1, 50))
  X <- rbind(replicate(3, rnorm(100, 0)),
             replicate(3, rnorm(50, 0.6)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)

  expect_warning(
    trimmed <- ebalance.trim(fit, max.weight = 3, print.level = 0),
    "Trimming halted"
  )
  expect_s3_class(trimmed, "ebalance.trim")
  expect_false(trimmed$trim.feasible)
  expect_true(all(is.finite(trimmed$w)))
})

test_that("ebalance.trim populates target.margins (regression test for 0.1-8 bug)", {
  set.seed(20260427L)
  treatment <- c(rep(0, 100), rep(1, 50))
  X <- rbind(replicate(3, rnorm(100, 0)),
             replicate(3, rnorm(50, 0.6)))
  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)

  trimmed <- ebalance.trim(fit, print.level = 0)
  expect_false(is.null(trimmed$target.margins))
  expect_identical(trimmed$target.margins, fit$target.margins)
})

test_that("ebalance.trim rejects non-ebalance input", {
  expect_error(ebalance.trim(list(a = 1)), "ebalance object")
})

# Internal helpers. Not exported.

# Cap-then-exponentiate. exp() of a double overflows at about 709.78;
# at that point the result is Inf, and Inf multiplied against a zero
# entry of the design matrix or a zero base weight yields NaN, which
# breaks the convergence check downstream.
#
# This helper caps the input at 700 before exponentiating. The cap
# is inactive (no-op) for any well-conditioned problem, since fitted
# coefficients on standardized covariates produce linear predictors
# in the single digits. It only kicks in when the optimizer is being
# pushed into the overflow regime, e.g. by an aggressive
# ebalance.trim() target. In that regime, returning a very large but
# finite weight lets line.searcher() and eb()'s gradient check
# continue numerically; the optimizer then steers back away from the
# cap on subsequent iterations.
.safe_exp <- function(x, cap = 700) {
  exp(pmin(x, cap))
}

# =============================================================================
# S3 methods for ebalance and ebalance.trim objects.
#
# All purely additive — old code that reads list fields by name continues
# to work; these methods just provide nicer defaults for print/summary/plot
# and a length-n weights() vector that's drop-in for lm()/svyglm().
# =============================================================================

# ---- internal helpers -------------------------------------------------------

# Build a balance table comparing pre-weighting and post-weighting moments
# for the columns of X under a binary Treatment vector and a vector of
# control weights w.controls (length = number of controls).
.balance_table <- function(Treatment, X, w.controls) {
  is.t <- Treatment == 1
  is.c <- Treatment == 0
  Xt   <- X[is.t, , drop = FALSE]
  Xc   <- X[is.c, , drop = FALSE]

  mean.t       <- colMeans(Xt)
  mean.c.pre   <- colMeans(Xc)
  mean.c.post  <- apply(Xc, 2, weighted.mean, w = w.controls)

  # Pooled SD across treatment and control (pre), used as the
  # denominator for both pre- and post-weighting standardized
  # differences so they are directly comparable.
  var.t  <- apply(Xt, 2, var)
  var.c  <- apply(Xc, 2, var)
  sd.pool <- sqrt((var.t + var.c) / 2)
  sd.pool[sd.pool == 0] <- NA_real_

  out <- data.frame(
    mean.Tr        = mean.t,
    mean.Co.pre    = mean.c.pre,
    mean.Co.post   = mean.c.post,
    diff.pre       = mean.t - mean.c.pre,
    diff.post      = mean.t - mean.c.post,
    std.diff.pre   = (mean.t - mean.c.pre)  / sd.pool,
    std.diff.post  = (mean.t - mean.c.post) / sd.pool,
    row.names = colnames(X),
    stringsAsFactors = FALSE
  )
  out
}

# ---- print methods ----------------------------------------------------------

print.ebalance <- function(x, ...) {
  cat("Entropy balancing\n")
  cat("-----------------\n")
  ntreated  <- if (!is.null(x$Treatment)) sum(x$Treatment == 1) else NA
  ncontrols <- length(x$w)
  cat(sprintf("Treated:    %d\n", ntreated))
  cat(sprintf("Controls:   %d (sum of weights = %.3f)\n",
              ncontrols, sum(x$w)))
  nmom <- length(x$target.margins) - 1L  # subtract the norm.constant entry
  cat(sprintf("Moments:    %d covariate moment(s) balanced\n", nmom))
  cat(sprintf("Converged:  %s   (max moment deviation = %.3g)\n",
              x$converged, x$maxdiff))
  cat("\nUse summary() for a balance table, weights() for the per-unit\n")
  cat("weight vector (treated units get weight 1), and plot() for a\n")
  cat("Love plot of standardized differences.\n")
  invisible(x)
}

print.ebalance.trim <- function(x, ...) {
  cat("Entropy balancing (trimmed weights)\n")
  cat("-----------------------------------\n")
  ntreated  <- if (!is.null(x$Treatment)) sum(x$Treatment == 1) else NA
  ncontrols <- length(x$w)
  cat(sprintf("Treated:        %d\n", ntreated))
  cat(sprintf("Controls:       %d (sum of weights = %.3f)\n",
              ncontrols, sum(x$w)))
  nmom <- length(x$target.margins) - 1L
  cat(sprintf("Moments:        %d covariate moment(s) balanced\n", nmom))
  cat(sprintf("Converged:      %s   (max moment deviation = %.3g)\n",
              x$converged, x$maxdiff))
  cat(sprintf("Trim feasible:  %s   (max weight ratio = %.3f)\n",
              x$trim.feasible, max(x$w) / mean(x$w)))
  if (isFALSE(x$trim.feasible)) {
    cat("\n  ! requested max.weight target was not achieved;\n",
        "  ! the most recent feasible fit is returned.\n", sep = "")
  }
  invisible(x)
}

# ---- summary methods --------------------------------------------------------

summary.ebalance <- function(object, ...) {
  if (is.null(object$Treatment) || is.null(object$X)) {
    stop("This ebalance object was fit before Treatment/X were stored ",
         "in the result; refit with the current package version to use ",
         "summary().")
  }
  bal <- .balance_table(object$Treatment, object$X, object$w)
  out <- list(
    call.info = list(n.treated  = sum(object$Treatment == 1),
                     n.controls = sum(object$Treatment == 0),
                     converged  = object$converged,
                     maxdiff    = object$maxdiff),
    balance   = bal
  )
  class(out) <- "summary.ebalance"
  out
}

summary.ebalance.trim <- function(object, ...) {
  if (is.null(object$Treatment) || is.null(object$X)) {
    stop("This ebalance.trim object was fit before Treatment/X were stored ",
         "in the result; refit with the current package version to use ",
         "summary().")
  }
  bal <- .balance_table(object$Treatment, object$X, object$w)
  out <- list(
    call.info = list(n.treated      = sum(object$Treatment == 1),
                     n.controls     = sum(object$Treatment == 0),
                     converged      = object$converged,
                     maxdiff        = object$maxdiff,
                     trim.feasible  = object$trim.feasible,
                     max.weight.ratio = max(object$w) / mean(object$w)),
    balance   = bal
  )
  class(out) <- "summary.ebalance.trim"
  out
}

print.summary.ebalance <- function(x, digits = 4, ...) {
  ci <- x$call.info
  cat("Entropy balancing summary\n")
  cat(sprintf("  Treated:   %d   Controls: %d   Converged: %s   max moment deviation: %.3g\n",
              ci$n.treated, ci$n.controls, ci$converged, ci$maxdiff))
  cat("\nBalance table (means and standardized differences):\n\n")
  print(round(x$balance, digits = digits))
  invisible(x)
}

print.summary.ebalance.trim <- function(x, digits = 4, ...) {
  ci <- x$call.info
  cat("Entropy balancing (trimmed) summary\n")
  cat(sprintf("  Treated: %d   Controls: %d   Converged: %s   max moment deviation: %.3g\n",
              ci$n.treated, ci$n.controls, ci$converged, ci$maxdiff))
  cat(sprintf("  Trim feasible: %s   max weight ratio: %.3f\n",
              ci$trim.feasible, ci$max.weight.ratio))
  cat("\nBalance table (means and standardized differences):\n\n")
  print(round(x$balance, digits = digits))
  invisible(x)
}

# ---- weights methods --------------------------------------------------------
#
# Returns a length-n vector aligned to the original Treatment/X. Treated
# units receive weight 1; control units receive their entropy-balancing
# weight. Suitable for passing directly to lm(..., weights = w),
# svyglm(...), etc.

weights.ebalance <- function(object, ...) {
  if (is.null(object$Treatment)) {
    stop("This ebalance object has no Treatment field; refit with the ",
         "current package version to use weights() at full length, or ",
         "read object$w for the controls-only vector.")
  }
  out <- numeric(length(object$Treatment))
  out[object$Treatment == 1] <- 1
  out[object$Treatment == 0] <- object$w
  out
}

weights.ebalance.trim <- function(object, ...) {
  if (is.null(object$Treatment)) {
    stop("This ebalance.trim object has no Treatment field; refit with ",
         "the current package version to use weights() at full length, or ",
         "read object$w for the controls-only vector.")
  }
  out <- numeric(length(object$Treatment))
  out[object$Treatment == 1] <- 1
  out[object$Treatment == 0] <- object$w
  out
}

# ---- plot methods -----------------------------------------------------------
#
# Love plot: absolute standardized differences pre- and post-weighting,
# one row per covariate. Base graphics, no ggplot2 dependency.

plot.ebalance <- function(x,
                          abs.values = TRUE,
                          xlab = NULL,
                          main = NULL,
                          ...) {
  if (is.null(xlab)) {
    xlab <- if (abs.values) "Absolute standardized difference"
            else "Standardized difference"
  }
  if (is.null(main)) {
    main <- "Covariate balance (before vs. after entropy balancing)"
  }
  if (is.null(x$Treatment) || is.null(x$X)) {
    stop("plot() requires the Treatment and X fields, which are stored ",
         "by the current package version. Refit to use plot().")
  }
  bal <- .balance_table(x$Treatment, x$X, x$w)
  pre  <- bal$std.diff.pre
  post <- bal$std.diff.post
  if (abs.values) { pre <- abs(pre); post <- abs(post) }
  k <- nrow(bal)
  ord <- order(abs(bal$std.diff.pre), decreasing = FALSE)
  ylim <- c(1, k)
  xlim <- range(c(0, pre, post), na.rm = TRUE)
  if (abs.values) xlim[1] <- 0

  op <- par(mar = c(4, max(8, max(nchar(rownames(bal))) * 0.6), 3, 1))
  on.exit(par(op), add = TRUE)
  plot(pre[ord], seq_len(k), type = "n",
       xlim = xlim, ylim = ylim,
       xlab = xlab, ylab = "", yaxt = "n", main = main, ...)
  axis(2, at = seq_len(k), labels = rownames(bal)[ord], las = 1)
  if (abs.values) abline(v = 0, col = "grey80")
  else            abline(v = 0, col = "grey80")
  points(pre[ord],  seq_len(k), pch = 1, col = "black")
  points(post[ord], seq_len(k), pch = 19, col = "darkblue")
  legend("topright",
         legend = c("before", "after"),
         pch = c(1, 19), col = c("black", "darkblue"),
         bty = "n", inset = 0.02)
  invisible(bal)
}

plot.ebalance.trim <- function(x, ...) plot.ebalance(x, ...)

baltest.collect <-
function(matchbal.out,var.names,after=TRUE)
 {
  storemat <- matrix(NA,length(var.names),10)
  colnames(storemat) <- c("mean.Tr","mean.Co","sdiff","sdiff.pooled","var.ratio","T pval","KS pval","qqmeandiff","qqmeddiff","qqmaxdiff")
  rownames(storemat) <- var.names

  if(after==FALSE){
    stopifnot(length(matchbal.out$BeforeMatching)==length(var.names))
    for(i in 1:length(matchbal.out$BeforeMatching))
     {
      storemat[i,1] <- matchbal.out$BeforeMatching[[i]]$mean.Tr
      storemat[i,2] <- matchbal.out$BeforeMatching[[i]]$mean.Co
      storemat[i,3] <- matchbal.out$BeforeMatching[[i]]$sdiff
      storemat[i,4] <- matchbal.out$BeforeMatching[[i]]$sdiff.pooled
      storemat[i,5] <- matchbal.out$BeforeMatching[[i]]$var.ratio
      storemat[i,6] <- matchbal.out$BeforeMatching[[i]]$p.value
      TT <- try(matchbal.out$BeforeMatching[[i]]$ks$ks.boot.pvalue,silent=TRUE)
      storemat[i,7] <- ifelse(is.null(TT),NA,TT)
      storemat[i,8] <- matchbal.out$BeforeMatching[[i]]$qqsummary$meandiff
      storemat[i,9] <- matchbal.out$BeforeMatching[[i]]$qqsummary$mediandiff
      storemat[i,10] <- matchbal.out$BeforeMatching[[i]]$qqsummary$maxdiff
     }

} else {
    stopifnot(length(matchbal.out$AfterMatching)==length(var.names))
    for(i in 1:length(matchbal.out$AfterMatching))
     {
      storemat[i,1] <- matchbal.out$AfterMatching[[i]]$mean.Tr
      storemat[i,2] <- matchbal.out$AfterMatching[[i]]$mean.Co
      storemat[i,3] <- matchbal.out$AfterMatching[[i]]$sdiff
      storemat[i,4] <- matchbal.out$AfterMatching[[i]]$sdiff.pooled
      storemat[i,5] <- matchbal.out$AfterMatching[[i]]$var.ratio
      storemat[i,6] <- matchbal.out$AfterMatching[[i]]$p.value
      TT <- try(matchbal.out$AfterMatching[[i]]$ks$ks.boot.pvalue,silent=TRUE)
      storemat[i,7] <- ifelse(is.null(TT),NA,TT)
      storemat[i,8] <- matchbal.out$AfterMatching[[i]]$qqsummary$meandiff
      storemat[i,9] <- matchbal.out$AfterMatching[[i]]$qqsummary$mediandiff
      storemat[i,10] <- matchbal.out$AfterMatching[[i]]$qqsummary$maxdiff
     }
}
return(storemat)
}

matrixmaker <-
function(mat) {
  k <- ncol(mat)
  obs <- nrow(mat)
  ncombs <- (k * (k + 1)) / 2
  out <- matrix(NA, obs, ncombs)
  out.names <- character(ncombs)
  count <- 0
  for (i in seq_len(k)) {
    for (j in seq_len(i)) {
      count <- count + 1
      out[, count] <- mat[, i] * mat[, j]
      out.names[count] <- paste(colnames(mat)[i], ".", colnames(mat)[j], sep = "")
    }
  }
  colnames(out) <- out.names
  cbind(mat, out)
}

test_that("matrixmaker returns expected structure", {
  set.seed(7L)
  m <- matrix(rnorm(20), 5, 4)
  colnames(m) <- c("a", "b", "c", "d")

  out <- matrixmaker(m)
  k <- 4L
  expect_equal(ncol(out), k + (k * (k + 1)) / 2)
  expect_false("dummy" %in% colnames(out))
  expect_equal(out[, "a.a"], m[, "a"] * m[, "a"])
  expect_equal(out[, "c.b"], m[, "c"] * m[, "b"])
})

test_that("getsquares squares only non-binary columns", {
  set.seed(5L)
  m <- cbind(
    cont = rnorm(20),
    binary = sample(0:1, 20, TRUE),
    ternary = sample(1:3, 20, TRUE)
  )

  out <- getsquares(m)
  expect_equal(ncol(out), 5L)
  expect_true("cont.2" %in% colnames(out))
  expect_false("binary.2" %in% colnames(out))
  expect_true("ternary.2" %in% colnames(out))
})

ebalance <- function(Treatment,
                     X = NULL,
                     base.weight = NULL,
                     norm.constant = NULL,
                     coefs = NULL,
                     max.iterations = 200,
                     constraint.tolerance = 1,
                     print.level = 0,
                     data = NULL,
                     ...) {

  # ---- formula interface ---------------------------------------------------
  # If the user passed a two-sided formula as the first argument, build
  # Treatment and X from formula + data and continue with the matrix
  # interface. This is implemented as an in-function dispatch rather than
  # via S3 (UseMethod) to avoid a CRAN R CMD check NOTE that triggers when
  # ebalance is a generic and ebalance.trim — a long-standing top-level
  # function whose name accidentally matches the <generic>.<class> pattern
  # — looks like a method whose signature does not match the generic.
  if (inherits(Treatment, "formula")) {
    formula <- Treatment
    if (is.null(data)) {
      stop("'data' is required when calling ebalance() with a formula")
    }
    if (length(formula) != 3L) {
      stop("formula must be two-sided, e.g. treat ~ x1 + x2")
    }

    mf <- model.frame(formula, data = data, na.action = na.pass)
    treat <- model.response(mf)
    if (is.null(treat)) {
      stop("formula has no response (left-hand side); expected treat ~ x1 + x2 + ...")
    }

    Xmat <- model.matrix(formula, data = data)
    if ("(Intercept)" %in% colnames(Xmat)) {
      Xmat <- Xmat[, colnames(Xmat) != "(Intercept)", drop = FALSE]
    }

    Treatment <- treat
    X <- Xmat
  }

  # ---- input checks --------------------------------------------------------
  if (is.null(X)) {
    stop("'X' is required when 'Treatment' is not a formula")
  }
  if (sum(Treatment != 1 & Treatment != 0) > 0) {
    stop("Treatment indicator ('Treatment') must be a logical variable, TRUE (1) or FALSE (0)")
  }
  if (var(Treatment) == 0) {
    stop("Treatment indicator ('Treatment') must contain both treatment and control observations")
  }

  Treatment.in <- Treatment
  Treatment <- as.numeric(Treatment)
  X <- as.matrix(X)

  if (sum(is.na(X)) > 0) {
    stop("X contains missing data")
  }
  if (sum(is.na(Treatment)) > 0) {
    stop("Treatment contains missing data")
  }
  if (length(Treatment) != nrow(X)) {
    stop("length(Treatment) != nrow(X)")
  }
  if (length(max.iterations) != 1) {
    stop("length(max.iterations) != 1")
  }
  if (length(constraint.tolerance) != 1) {
    stop("length(constraint.tolerance) != 1")
  }

  # ---- setup ---------------------------------------------------------------
  ntreated  <- sum(Treatment == 1)
  ncontrols <- sum(Treatment == 0)

  if (is.null(base.weight)) {
    base.weight <- rep(1, ncontrols)
  }
  if (length(base.weight) != ncontrols) {
    stop("length(base.weight) !=  number of controls  sum(Treatment==0)")
  }

  co.x <- X[Treatment == 0, , drop = FALSE]
  co.x <- cbind(rep(1, ncontrols), co.x)

  if (qr(co.x)$rank != ncol(co.x)) {
    stop("collinearity in covariate matrix for controls (remove collinear covariates)")
  }

  tr.total <- colSums(X[Treatment == 1, , drop = FALSE])

  if (is.null(norm.constant)) {
    norm.constant <- ntreated
  }
  if (length(norm.constant) != 1) {
    stop("length(norm.constant) != 1")
  }

  tr.total <- c(norm.constant, tr.total)

  if (is.null(coefs)) {
    coefs <- c(log(tr.total[1] / sum(base.weight)), rep(0, (ncol(co.x) - 1)))
  }

  if (length(coefs) != ncol(co.x)) {
    stop("coefs needs to have same length as number of covariates plus one")
  }

  # ---- run algorithm -------------------------------------------------------
  eb.out <- eb(tr.total = tr.total,
               co.x = co.x,
               coefs = coefs,
               base.weight = base.weight,
               max.iterations = max.iterations,
               constraint.tolerance = constraint.tolerance,
               print.level = print.level)

  if (eb.out$converged && print.level > 0) {
    cat("Converged within tolerance \n")
  }

  z <- list(
    target.margins       = tr.total,
    co.xdata             = co.x,
    w                    = eb.out$Weights.ebal,
    coefs                = eb.out$coefs,
    maxdiff              = eb.out$maxdiff,
    norm.constant        = norm.constant,
    constraint.tolerance = constraint.tolerance,
    max.iterations       = max.iterations,
    base.weight          = base.weight,
    print.level          = print.level,
    converged            = eb.out$converged,
    Treatment            = Treatment.in,
    X                    = X
  )

  class(z) <- "ebalance"
  z
}

test_that("formula and matrix interfaces give identical results", {
  set.seed(20260427L)
  treatment <- c(rep(0, 50), rep(1, 30))
  X <- rbind(replicate(3, rnorm(50, 0)), replicate(3, rnorm(30, 0.5)))
  colnames(X) <- paste0("x", 1:3)
  df <- data.frame(treat = treatment, X)

  fit_mat <- ebalance(Treatment = treatment, X = X, print.level = 0)
  fit_for <- ebalance(treat ~ x1 + x2 + x3, data = df, print.level = 0)

  expect_s3_class(fit_for, "ebalance")
  expect_equal(fit_for$w, fit_mat$w, tolerance = 1e-12)
  expect_equal(fit_for$coefs, fit_mat$coefs, tolerance = 1e-12)
})

test_that("formula method requires data argument", {
  expect_error(ebalance(treat ~ x1, data = NULL), "'data' is required")
})

test_that("formula method rejects one-sided formula", {
  df <- data.frame(treat = c(0, 1), x1 = c(0, 1))
  expect_error(ebalance(~ x1, data = df))
})

test_that("formula method drops the intercept column from X", {
  set.seed(20260427L)
  df <- data.frame(
    treat = c(rep(0, 50), rep(1, 30)),
    x1 = rnorm(80), x2 = rnorm(80), x3 = rnorm(80)
  )
  fit <- ebalance(treat ~ x1 + x2 + x3, data = df, print.level = 0)
  # 1 (intercept moment) + 3 covariate moments = 4 target margins
  expect_length(fit$target.margins, 4L)
  expect_false("(Intercept)" %in% colnames(fit$X))
})

test_that("ebalance() with matrix interface converges on a simple toy", {
  set.seed(20260427L)
  treatment <- c(rep(0, 50), rep(1, 30))
  X <- rbind(replicate(3, rnorm(50, 0)), replicate(3, rnorm(30, 0.5)))
  colnames(X) <- paste0("x", 1:3)

  fit <- ebalance(Treatment = treatment, X = X,
                  constraint.tolerance = 1e-8, print.level = 0)

  expect_s3_class(fit, "ebalance")
  expect_true(fit$converged)
  expect_length(fit$w, 50L)
  # post-weighting control means match treated means up to the
  # solver tolerance scaled by 1/ntreated
  ntreated <- sum(treatment == 1)
  for (j in seq_len(ncol(X))) {
    expect_equal(weighted.mean(X[treatment == 0, j], w = fit$w),
                 mean(X[treatment == 1, j]),
                 tolerance = 1e-8 / ntreated * 10)  # generous slack
  }
})

test_that("ebalance() handles a single covariate without dim drop", {
  set.seed(20260427L)
  treatment <- c(rep(0, 60), rep(1, 40))
  X <- matrix(c(rnorm(60), rnorm(40, 0.7)), ncol = 1,
              dimnames = list(NULL, "x1"))

  fit <- ebalance(Treatment = treatment, X = X, print.level = 0)
  expect_true(fit$converged)
  expect_length(fit$w, 60L)
})

test_that("ebalance() supports non-uniform base.weight", {
  set.seed(20260427L)
  treatment <- c(rep(0, 80), rep(1, 40))
  X <- rbind(replicate(4, rnorm(80, 0)), replicate(4, rnorm(40, 0.4)))
  ncontrols <- sum(treatment == 0)
  bw <- runif(ncontrols, 0.5, 1.5)

  fit <- ebalance(Treatment = treatment, X = X, base.weight = bw,
                  print.level = 0)
  expect_true(fit$converged)
})

test_that("ebalance() rejects malformed inputs with clear messages", {
  X <- matrix(rnorm(20), 10, 2)
  treatment <- c(rep(0, 5), rep(1, 5))

  expect_error(ebalance(Treatment = c(0, 0, 0, 2), X = X[1:4, ]),
               "logical")
  expect_error(ebalance(Treatment = rep(0, 10), X = X),
               "treatment and control")
  bad_X <- X
  bad_X[1, 1] <- NA
  expect_error(ebalance(Treatment = treatment, X = bad_X),
               "missing data")
})

getsquares <-
function(mat) {
  mat.add <- matrix(NA, nrow(mat), 0)
  mat.add.names <- c()
  for (i in seq_len(ncol(mat))) {
    if (length(unique(mat[, i])) > 2) {
      mat.add <- cbind(mat.add, mat[, i]^2)
      mat.add.names <- c(paste(colnames(mat)[i], ".2", sep = ""), mat.add.names)
    }
  }
  colnames(mat.add) <- mat.add.names
  out <- cbind(mat, mat.add)
  out
}

\name{matrixmaker}
\alias{matrixmaker}


\title{
Generate Matrix of One-way Interactions and Squared Terms

}
\description{
Takes a matrix of covariates and generates a new matrix that contains the original covariates, all one-way interaction terms, and all squared terms.
}
\usage{
matrixmaker(mat)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{mat}{n by k numeric matrix of covariates.
}
}

\value{
n by (k*(k+1))/2 +1) matrix of covariates with original covariates, all one-way interaction terms, and all squared terms. 
}

\author{
Jens Hainmueller
}

\seealso{
See \code{\link{getsquares}}
}

\examples{

# create toy matrix
mold <- replicate(3,rnorm(50))
colnames(mold) <- paste("x",1:3,sep="")
head(mold)
# create new matrix
mnew <- matrixmaker(mold)
head(mnew)

}

\name{ebalance-methods}
\alias{print.ebalance}
\alias{print.ebalance.trim}
\alias{summary.ebalance}
\alias{summary.ebalance.trim}
\alias{print.summary.ebalance}
\alias{print.summary.ebalance.trim}
\alias{plot.ebalance}
\alias{plot.ebalance.trim}
\alias{weights.ebalance}
\alias{weights.ebalance.trim}

\title{Methods for ebalance and ebalance.trim objects}

\description{
Convenience methods for inspecting and using objects returned by
\code{\link{ebalance}} and \code{\link{ebalance.trim}}.
}

\usage{
\method{print}{ebalance}(x, \dots)
\method{print}{ebalance.trim}(x, \dots)
\method{summary}{ebalance}(object, \dots)
\method{summary}{ebalance.trim}(object, \dots)
\method{print}{summary.ebalance}(x, digits = 4, \dots)
\method{print}{summary.ebalance.trim}(x, digits = 4, \dots)
\method{plot}{ebalance}(x, abs.values = TRUE,
     xlab = NULL, main = NULL, \dots)
\method{plot}{ebalance.trim}(x, \dots)
\method{weights}{ebalance}(object, \dots)
\method{weights}{ebalance.trim}(object, \dots)
}

\arguments{
  \item{x, object}{An object of class \code{ebalance} or \code{ebalance.trim}.}
  \item{abs.values}{Logical. If \code{TRUE} (default) the Love plot shows
    absolute standardized differences; if \code{FALSE} signed.}
  \item{xlab, main}{Standard graphical arguments passed to \code{plot()}.}
  \item{digits}{Number of digits used when printing the summary table.}
  \item{\dots}{Additional arguments. Currently unused for \code{print},
    \code{summary}, \code{weights}; passed to \code{plot()} for plot methods.}
}

\details{
\code{print} gives a one-screen overview: counts of treated/control units,
number of moments balanced, convergence status, and (for trimmed objects)
whether the trim target was met.

\code{summary} returns a list of class \code{summary.ebalance} (or
\code{summary.ebalance.trim}) containing a balance table that compares
treated and control covariate means before and after weighting along with
the corresponding standardized differences.

\code{plot} produces a Love plot of the standardized differences before and
after weighting, one row per covariate.

\code{weights} returns a length-\eqn{n} numeric vector aligned to the
original \code{Treatment}: treated observations receive weight 1 and control
observations receive their entropy-balancing weight. This is suitable for
use with \code{lm(\dots, weights = w)} and other model fitters that accept
case weights.
}

\value{
\code{print} and the print methods for summary objects return their input
invisibly. \code{summary} returns an object of class \code{summary.ebalance}
or \code{summary.ebalance.trim} containing \code{$call.info} and
\code{$balance}. \code{plot} returns the balance table invisibly. \code{weights}
returns a numeric vector of length equal to the original
\code{Treatment} vector.
}

\seealso{
\code{\link{ebalance}}, \code{\link{ebalance.trim}}.
}

\examples{
set.seed(1)
df <- data.frame(
  treat = c(rep(1, 30), rep(0, 50)),
  x1    = c(rnorm(30, 0.5), rnorm(50, 0)),
  x2    = c(rnorm(30, 0.5), rnorm(50, 0)),
  x3    = c(rnorm(30, 0.5), rnorm(50, 0))
)

fit <- ebalance(treat ~ x1 + x2 + x3, data = df)

# print(): one-screen overview of the fit
print(fit)

# summary(): pre/post means and standardized differences for each
# covariate; the post-weighting std diffs should be near zero.
summary(fit)

# weights(): length-n vector aligned to the original treatment.
# Treated observations get weight 1; control observations get the
# entropy-balancing weight. Drop-in for lm(..., weights = w).
w <- weights(fit)
length(w) == nrow(df)
all(w[df$treat == 1] == 1)

# Same methods on a trimmed object
trimmed <- ebalance.trim(fit)
print(trimmed)              # also shows trim.feasible
summary(trimmed)
weights(trimmed)[1:5]

\dontrun{
# Love plot of standardized differences before vs. after
plot(fit)
plot(trimmed)
}
}

\name{ebalance.trim}
\alias{ebalance.trim}
\title{
Trimming of Weights for Entropy Balancing
}
\description{
Trim weights obtained from entropy balancing. Takes the output from a call to
\code{\link{ebalance}} and trims the weights (subject to the moment conditions)
so that the ratio of the maximum (or minimum) weight to the mean weight is
reduced to satisfy a user-specified target. If no target is specified, the
maximum weight ratio is automatically trimmed as far as is feasible given the
data.
}
\usage{
ebalance.trim(ebalanceobj, max.weight = NULL,
              min.weight = 0, max.trim.iterations = 200,
              max.weight.increment = 0.92,
              min.weight.increment = 1.08,
              print.level = 0)
}

\arguments{
  \item{ebalanceobj}{
An object from a call to \code{ebalance}.
}
  \item{max.weight}{
Optional target for the ratio of the maximum to mean weight.
}
  \item{min.weight}{
Optional target for the ratio of the minimum to mean weight.
}         
  \item{max.trim.iterations}{
Maximum number of trimming iterations.
}
  \item{max.weight.increment}{
Increment for iterative trimming of the ratio of the maximum to mean weight (a scalar between 0-1, .92 indicates that 
the attempted reduction in the max ratio is 8 percent). 
}
  \item{min.weight.increment}{
Increment for iterative trimming of the ratio of the minimum to mean weight (a scalar  > 1, 1.08 indicates that 
the attempted reduction in the max ratio is 8 percent).
}
  \item{print.level}{
Controls the level of printing: 0 (silent, the default), 1 (normal printing), 2 (detailed), and 3 (very detailed).
}
}

\value{
An list object of class \code{ebalance.trim} with the following elements:
  \item{target.margins}{A vector that contains the target moments coded from the covariate distributions of the treatment group.}
  \item{co.xdata}{A matrix that contains the covariate data from the control group.}
  \item{w}{A vector that contains the control group weights assigned by trimming entropy balancing algorithm.}
  \item{coefs }{A vector that contains coefficients from the reweighting algorithm.}
  \item{maxdiff }{A scalar that contains the maximum deviation between the moments of the reweighted data and the target moments.}
  \item{norm.constant}{Normalizing constant used.}
  \item{constraint.tolerance}{The tolerance level used for the balance constraints.}
  \item{max.iterations}{Maximum number of trimming iterations used.}
  \item{base.weight}{The base weight used.}
  \item{converged}{Logical flag if the inner entropy-balancing algorithm converged within tolerance on the last successful iteration.}
  \item{trim.feasible}{Logical flag indicating whether the requested trimming target was achieved. \code{TRUE} when (a) an explicit \code{max.weight} (and optional \code{min.weight}) target was met, or (b) automated minimization mode finished. \code{FALSE} when an explicit target was not met because the maximum number of iterations was exceeded or the inner solve became numerically singular; in that case a warning is emitted and the most recent feasible fit is returned.}
}

\references{

Hainmueller, J. (2012) 'Entropy Balancing for Causal Effects: A Multivariate Reweighting Method to Produce Balanced Samples in Observational Studies', Political Analysis (Winter 2012) 20 (1): 25--46.

Zaslavsky, A. (1988), 'Representing local reweighting area adjustments by of households', Survey Methodology 14(2), 265--288.

Ireland, C. and Kullback, S. (1968), 'Contingency tables with given marginals', Biometrika 55, 179--188.

Kullback, S. (1959), Information Theory and Statistics, Wiley, NY.

}
\author{
Jens Hainmueller
}

\seealso{
Also see \code{\link{ebalance}}.
}

\examples{

# Toy data with substantial covariate imbalance
set.seed(20260427)
n_t <- 50; n_c <- 100
df  <- data.frame(
  treat = c(rep(1, n_t), rep(0, n_c)),
  x1    = c(rnorm(n_t, 0.6), rnorm(n_c, 0)),
  x2    = c(rnorm(n_t, 0.6), rnorm(n_c, 0)),
  x3    = c(rnorm(n_t, 0.6), rnorm(n_c, 0))
)
fit <- ebalance(treat ~ x1 + x2 + x3, data = df)

# ---- Auto-minimization mode ---------------------------------------
# Without a target, ebalance.trim() iteratively reduces the maximum
# weight ratio as far as the data allows. trim.feasible is TRUE by
# definition for auto mode.
trimmed <- ebalance.trim(fit)
trimmed                        # print method shows trim.feasible + max ratio
summary(trimmed)               # balance table for the trimmed weights

# Compare untrimmed vs. trimmed weight distributions
round(summary(fit$w),     2)
round(summary(trimmed$w), 2)

# ---- Explicit max.weight target -----------------------------------
# Pick a target above the natural minimum ratio so it's achievable.
target <- max(fit$w / mean(fit$w)) * 1.5
trimmed2 <- ebalance.trim(fit, max.weight = target)
trimmed2$trim.feasible         # TRUE — target was met

# ---- Infeasible target: graceful fallback (new in 0.2.0) ----------
# Asking for something the data cannot support no longer crashes.
# A warning is emitted and the most recent feasible fit is returned
# with trim.feasible = FALSE.
trimmed3 <- suppressWarnings(ebalance.trim(fit, max.weight = 1.2))
trimmed3$trim.feasible         # FALSE — target was infeasible
max(trimmed3$w) / mean(trimmed3$w)   # the best we could do

# ---- Use the trimmed weights downstream ---------------------------
df$y <- df$treat * 5 + df$x1 + df$x2 + df$x3 + rnorm(nrow(df))
df$w <- weights(trimmed)       # length = nrow(df), treated = 1
coef(lm(y ~ treat, data = df, weights = w))["treat"]

}

\name{line.searcher}
\alias{line.searcher}
\title{
Optimal step length search for entropy balancing algorithm 
}
\description{
Function called internally by \code{ebalance} and \code{ebalance.trim} to compute optimal step length for entropy balancing algorithm. This function would normally not be called manually by a user.  
}
\usage{
line.searcher(Base.weight, Co.x, 
Tr.total, coefs, Newton, ss)
}
\arguments{
  \item{Base.weight}{
NA
}
  \item{Co.x}{
NA
}
  \item{Tr.total}{
NA
}
  \item{coefs}{
NA
}
  \item{Newton}{
NA
}
  \item{ss}{
NA
}
}

\value{
A list with the results from the search.
}

\author{
Jens Hainmueller
}

\seealso{
\code{ebalance}, \code{ebalance.trim}
}
\examples{
##---- NA -----
}

\name{baltest.collect}
\alias{baltest.collect}
\title{
Collect Covariate Balance Statistics 
}
\description{
A function that summarizes the covariate balance statistics that are computed by \code{MatchBalance(Matching)} in a balance table.  
}
\usage{
baltest.collect(matchbal.out, var.names, after = TRUE)
}

\arguments{
  \item{matchbal.out}{
An object from a call to \code{MatchBalance(Matching)}
}
  \item{var.names}{
A vector of covariate names.
}
  \item{after}{
A logical flag for whether the results from before or after Matching should be summarized. If TRUE \code{baltest.collect} summarizes the results from the covariate balance checks that \code{MatchBalance} computes in the matched data. If FALSE the results from the balance checks in the unmatched data are used.
}
}
\details{
See \code{MatchBalance(Matching)} for details.
}
\value{
  A matrix that contains the covariate balance statistics in tabular format.
}

\author{
Jens Hainmueller
}


\seealso{
\code{MatchBalance} in the \code{Matching} package.
}
\examples{

## load(Matching) to run this example
## create toy data: one treatment indicator and three covariates X1-3
#dat <- data.frame(treatment=rbinom(50,size=1,prob=.5),replicate(3,rnorm(50)))
#covarsname <- colnames(dat)[-1]

## run balance checks
#mout <- MatchBalance(treatment~X1+X2+X3,data=dat)

## summarize in balance table
#baltest.collect(matchbal.out=mout,var.names=covarsname,after=FALSE)

}

\name{ebalance}
\alias{ebalance}

\title{
Entropy balancing
}
\description{
This function implements entropy balancing, a data preprocessing procedure that
allows users to reweight a dataset. The preprocessing is based on a maximum
entropy reweighting scheme that assigns weights to each unit such that the
covariate distributions in the reweighted data satisfy a set of moment
conditions specified by the researcher. This can be useful to balance covariate
distributions in observational studies with a binary treatment where the
control group data can be reweighted to match the covariate moments in the
treatment group. Entropy balancing can also be used to reweight a survey sample
to known characteristics from a target population. The weights that result
from entropy balancing can be passed to regression or other models to
subsequently analyze the reweighted data.

By default, \code{ebalance} reweights the covariate distributions from a
control group to match target moments computed from a treatment group such
that the reweighted data can be used to analyze the average treatment effect
on the treated.

Two interfaces are supported. With \code{Treatment} as a numeric or logical
vector, supply the covariate matrix \code{X} directly. With \code{Treatment}
as a two-sided \code{formula}, supply a \code{data} frame; the formula's
left-hand side is used as the treatment indicator and the right-hand side
as the covariate matrix (the intercept column is dropped automatically).
}

\usage{
ebalance(Treatment, X = NULL, base.weight = NULL,
         norm.constant = NULL, coefs = NULL,
         max.iterations = 200, constraint.tolerance = 1,
         print.level = 0, data = NULL, \dots)
}

\arguments{
  \item{Treatment}{
For the default method: a vector indicating the observations to reweight
(controls) and those used to compute target moments (treatment). This can be
a logical vector or a numeric vector where 0 denotes control observations
and 1 denotes treatment observations. For the formula method: a two-sided
formula of the form \code{treat ~ x1 + x2 + ...}, with the treatment
indicator on the left-hand side and the covariates on the right.
}
  \item{X}{
A matrix containing the covariates to include in the reweighting. To adjust
the means of the covariates, include the raw covariates. To adjust the
variances, include squared terms; for co-moments, include interaction terms.
All columns must have positive variance and the matrix must be invertible.
No missing data is allowed.
}
  \item{data}{
For the formula method: a data frame containing the variables in
\code{Treatment}.
}
  \item{base.weight}{
An optional vector of base weights for the maximum entropy reweighting (one
weight per control unit). Default: uniform base weights.
}
  \item{norm.constant}{
An optional normalizing constant. By default the weights are normalized such
that their sum equals the number of treated observations.
}
  \item{coefs}{
An optional vector of starting coefficients.
}
  \item{max.iterations}{
Maximum number of iterations.
}
  \item{constraint.tolerance}{
Tolerance for declaring the moments in the reweighted data equal to the
target moments.
}
  \item{print.level}{
Controls the level of printing: 0 (silent, the default), 1 (normal printing),
2 (detailed), and 3 (very detailed).
}
  \item{\dots}{
Additional arguments. For the formula method, passed through to the default
method.
}
}

\value{
A list of class \code{ebalance} with the following elements:
  \item{target.margins}{Target moments computed from the treatment group.}
  \item{co.xdata}{Covariate data from the control group (with intercept column).}
  \item{w}{Control-group weights assigned by entropy balancing (length = number of controls).}
  \item{coefs}{Coefficients from the reweighting algorithm.}
  \item{maxdiff}{Maximum deviation between reweighted moments and targets.}
  \item{norm.constant}{Normalizing constant used.}
  \item{constraint.tolerance}{Tolerance level used for the balance constraints.}
  \item{max.iterations}{Maximum number of iterations used.}
  \item{base.weight}{Base weight used.}
  \item{print.level}{Print level used.}
  \item{converged}{Logical flag indicating convergence within tolerance.}
  \item{Treatment}{The treatment indicator vector as supplied (length = number of observations).}
  \item{X}{The covariate matrix as supplied.}
}

\references{

Hainmueller, J. (2012) 'Entropy Balancing for Causal Effects: A Multivariate Reweighting Method to Produce Balanced Samples in Observational Studies', Political Analysis (Winter 2012) 20 (1): 25--46.

Zaslavsky, A. (1988), 'Representing local reweighting area adjustments by of households', Survey Methodology 14(2), 265--288.

Ireland, C. and Kullback, S. (1968), 'Contingency tables with given marginals', Biometrika 55, 179--188.

Kullback, S. (1959), Information Theory and Statistics, Wiley, NY.

}

\author{Jens Hainmueller}

\seealso{
\code{\link{ebalance.trim}} for trimming weights, and the
\code{\link[=summary.ebalance]{summary}}, \code{\link[=plot.ebalance]{plot}}, and
\code{\link[=weights.ebalance]{weights}} methods for inspection and downstream use.
}

\examples{

# Toy observational-study data: treatment is associated with older,
# more educated, higher-income units; the true effect on the outcome
# is 5, but a naive comparison is biased upward by the confounders.
set.seed(42)
n_t <- 75; n_c <- 250
df  <- data.frame(
  treat  = c(rep(1, n_t), rep(0, n_c)),
  age    = c(rnorm(n_t, 45,  8), rnorm(n_c, 38, 10)),
  educ   = c(rnorm(n_t, 16,  2.5), rnorm(n_c, 13, 3)),
  income = c(rnorm(n_t, 65, 12), rnorm(n_c, 50, 15))
)
df$y <- 0.1 * df$age + 0.3 * df$educ + 0.05 * df$income +
        5 * df$treat + rnorm(nrow(df), 0, 3)

# ---- Naive (biased) regression ------------------------------------
coef(lm(y ~ treat, data = df))["treat"]   # ATT estimate; pulled up by confounders

# ---- Entropy balancing: formula interface -------------------------
fit <- ebalance(treat ~ age + educ + income, data = df)
fit                       # one-screen overview via print()
summary(fit)              # balance table: pre/post means and std diffs

# ---- Equivalent matrix interface ----------------------------------
X <- as.matrix(df[, c("age", "educ", "income")])
fit2 <- ebalance(Treatment = df$treat, X = X)
all.equal(fit$w, fit2$w)  # identical results

# ---- Use the weights downstream ----------------------------------
df$w <- weights(fit)              # length = nrow(df); treated get 1
coef(lm(y ~ treat, data = df,     # weighted regression, ATT
        weights = w))["treat"]

# ---- Visualize balance --------------------------------------------
\dontrun{
plot(fit)                         # base-R Love plot, no dependencies
}

}

\name{eb}
\alias{eb}
\title{
Function for Entropy Balancing
}
\description{
This function is called internally by \code{\link{ebalance}} and \code{\link{ebalance.trim}} to implement entropy balancing. This function would normally not be called manually by a user.
}
\usage{
eb(tr.total = tr.total, co.x = co.x,
   coefs = coefs, base.weight = base.weight, 
   max.iterations = max.iterations, 
   constraint.tolerance = constraint.tolerance, 
   print.level = print.level)
}

\arguments{
  \item{tr.total}{
NA
}
  \item{co.x}{
NA
}
  \item{coefs}{
NA
}
  \item{base.weight}{
NA
}
  \item{max.iterations}{
NA
}
  \item{constraint.tolerance}{
NA
}
  \item{print.level}{
NA
}
}

\value{ A list containing the results from the algorithm.
}

\author{
Jens Hainmueller
}

\seealso{
\code{ebalance}, \code{ebalance.trim}
}
\examples{
##---- NA -----
}

\name{getsquares}
\alias{getsquares}

\title{
Generate Matrix of Squared Terms
}
\description{
Takes a matrix of covariates and generates a new matrix that contains the original covariates and all squared terms. Squared terms for binary covariates are omitted.
}
\usage{
getsquares(mat)
}
\arguments{
  \item{mat}{n by k numeric matrix of covariates.
  }
}

\value{
n by k*2 numeric matrix that contains the original covariates plus all squared terms.
}

\author{
Jens Hainmueller
}

\seealso{
See \code{\link{matrixmaker}}
}
\examples{
# create toy matrix
mold <- replicate(3,rnorm(50))
colnames(mold) <- paste("x",1:3,sep="")
head(mold)
# create new matrix
mnew <- getsquares(mold)
head(mnew)
}
